"""
    This file contains the functions HaarTransform, InvHaarTransform, and
    HaarCoefficients that perform discrete wavelet transormations on a given
    input array.
"""

import numpy as np

__all__ = ['HaarTransform', 'InvHaarTransform', 'HaarCoefficients', 'wh_alj', 'wh_clj']

def HaarTransform(Input_Array, Normalize=False):
    """
        Compute the discrete Haar wavelet transform of input_array.
        Parameters
        ----------
        Array : array_like
            Array to do wavelet transformation on.
        Normalize : bool
            Boolean for normalizing the coefficients or not.
            False by default.
        Returns
        -------
        WaveDec : array_like
            Coefficients of the discrete Haar wavelet transformation.
    """
    Array = np.array(Input_Array)
    nbins = len(Array)
    level = int(np.ceil(np.log2(nbins)))
        
    # -- Pad input_array with zeros if its length is not a power of two --
    if 2**level-nbins!=0:
        Array = np.append(Array, np.zeros(2**level-nbins))
    
    WaveDec = [] # Wavelet decomposition
    for l in range(1, level+1):
        J = 2**(level-l) # Number of bins per level
        C_l = np.array([wh_clj(Array, l, j) for j in range(1, J+1)])
        if Normalize==True:
            C_l = np.multiply(C_l, pow(2.0, -0.5*(l)))
        WaveDec.append(C_l)

    A_level = np.array([wh_alj(Array, level, 1)])
    if Normalize==True:
        A_level = np.multiply(A_level, pow(2.0, -0.5*level))
    WaveDec.append(A_level)

    return WaveDec


def HaarCoefficients(Input_Array, Normalize=False):
    """
        Compute all the coefficients of the discrete Haar wavelet
        transform of input_array.
        Parameters
        ----------
        Array : array_like
        Array to do wavelet transformation on.
        Normalize : bool
        Boolean for normalizing the coefficients or not.
        False by default.
        Returns
        -------
        (Ctype, Atype) : (array_like, array_like)
        Ctype and Atype coefficients of the discrete Haar wavelet transformation.
        """
    Array = np.array(Input_Array)
    nbins = len(Array)
    level = int(np.ceil(np.log2(nbins)))
    
    # -- Pad input_array with zeros if its length is not a power of two --
    if 2**level-nbins!=0:
        Array = np.append(Array, np.zeros(2**level-nbins))
    
    Ctype = [] # Ctype coefficients
    Atype = [] # Atype coefficients
    for l in range(1, level+1):
        J = 2**(level-l) # Number of bins per level
        C = np.array([wh_clj(Array, l, j) for j in range(1, J+1)])
        A = np.array([wh_alj(Array, l, j) for j in range(1, J+1)])
        if Normalize==True:
            C = np.multiply(C, pow(2.0, -0.5*l))
            A = np.multiply(A, pow(2.0, -0.5*l))
        Ctype.append(C)
        Atype.append(A)
            
    return (Ctype, Atype)

def InvHaarTransform(wavelet_default, normalize=False):
    N_levels = len(wavelet_default)
    lengthperlevel = []
    for level in wavelet_default:
        lsize = len(level)
        lengthperlevel.append(lsize)
    Nbins = sum(lengthperlevel)
    Lmax = N_levels - 1
    
    signal = [0]*Nbins
    for L in range(1,N_levels):
        if normalize:
            norm = 1 #factors already taken care of in forward transformation
        else:
            norm = pow(2, -0.5*L) #forward transformation didn't add this factor
        for j in range(1, lengthperlevel[L-1]+1):
            coeff = norm*wavelet_default[L-1][j-1]
            newpart = np.multiply(coeff, h1_wavelet(L, j, Nbins=Nbins))
            signal = np.add(signal, newpart)
    #Add A0/h0:
    if normalize:
        norm = 1 #factors already taken care of in forward transformation
    else:
        norm = pow(2, -0.5*Lmax) #forward transformation didn't add this factor
    coeffA0 = norm*wavelet_default[-1][0]
    lastpart = np.multiply(coeffA0, h0_wavelet(Lmax, 1, Nbins=Nbins))
    signal = np.add(signal, lastpart)
    return signal


def h0_wavelet(L, j, Nbins=128):
    norm = pow(2,-0.5*L)
    baselength = 2**L
    numberofj = int(Nbins/baselength)
    null = [0]*baselength
    waveform = [1*norm]*baselength
    return (j-1)*null + waveform + (numberofj - j)*null

def h1_wavelet(L, j, Nbins=128):
    norm = pow(2,-0.5*L)
    baselength = 2**L
    halflength = 2**(L-1)
    numberofj = int(Nbins/baselength)
    null = [0]*baselength
    waveform = [1*norm]*halflength + [-1*norm]*halflength
    return (j - 1)*null + waveform + (numberofj - j)*null

def wh_alj(Array, l, j):
    i_initial = (2**l)*(j-1)+1
    i_final = (2**l)*j
    Alj = 0
    for i in range(i_initial, i_final+1):
        Alj += Array[i-1]
    return Alj

def wh_clj(Array, l, j):
    Clj = wh_alj(Array, l=l-1, j=2*j-1)-wh_alj(Array, l=l-1, j=2*j)
    return Clj




"""
Kawking
"""
from __future__ import absolute_import

from .scalograms import *
from .nsigmaplots import *
from .plottingtools import *

__all__ = scalograms.__all__ + nsigmaplots.__all__ + plottingtools.__all__



"""
This file contains the NsetsMethod class that performs the
'nsets' method called by waveletanalysis.py
"""
from __future__ import absolute_import
import numpy as np
import scipy.special as spf
from scipy.optimize import curve_fit
import math
from mpmath import mp
from ..w_transform import HaarTransform
from .analysistools import *

g_nsigmamax = 10
g_pmin = mp.erfc(g_nsigmamax * mp.sqrt(0.5))
g_digits = int(math.ceil(-1 * mp.log10(g_pmin)))
g_logdigits = 17 #for nsigma and log10p output
mp.dps = g_digits + 2 #a bit extra

__all__ = ['NsetsMethod']

class NsetsMethod:
    """
        NsetsMethod class contains all functions and instances used in the 'nsets'
        method.
    """
    def __init__(self, data, hypothesis, nsets, extrapolate=False, fastGaussian=False, seed=123):
        """
        Parameters
        ----------
        data : array_like
        hypothesis : array_like
        nsets : int
        extrapolate : bool
        fastGaussian : bool
        seed : int
        """
        self.WaveDec_data = HaarTransform(data, Normalize=False)
        self.WaveDec_hypo = HaarTransform(hypothesis, Normalize=False)
        
        len_wdata = len(self.WaveDec_data)
        len_whypo = len(self.WaveDec_hypo)
        assert(len_wdata==len_whypo), "Data and hypothesis must have the same wavelet decomposition level."
        
        self.Level = len_wdata-1
        self.Seed = seed
        self.Nsets = nsets
        self.fast = fastGaussian

        WaveDec_nsets = np.empty((nsets), dtype=object) # Wavelet dec of the nset pseudodata
        WaveDec_nsets[0] = self.WaveDec_data
        for i in range(1, nsets):
            pseudodata = self.GeneratePoisson(hypothesis, (self.Seed*i))
            WaveDec_nsets[i] = HaarTransform(pseudodata, Normalize=False)
        
        PseudoWD_PerBin = _empty_like(self.WaveDec_data)
        for l in range(self.Level):
            J = 2**(self.Level-l-1)
            for j in range(J):
                Ccoeff_list = [WaveDec_nsets[i][l][j] for i in range(nsets)]
                unique, counts = np.unique(Ccoeff_list, return_counts=True)
                PseudoWD_PerBin[l][j] = [unique, counts]
            
        Acoeff_list = [WaveDec_nsets[i][self.Level][0] for i in range(nsets)]
        unique, counts = np.unique(Acoeff_list, return_counts=True)
        PseudoWD_PerBin[-1][0] = [unique, counts]
        
        self.Histogram = PseudoWD_PerBin
        self.zipHistogram = self.zipHistogram(self.Histogram)
        
        if self.fast==True:
            nsigmafit = _zeros_like(self.WaveDec_data)
            for l, level in enumerate(self.Histogram):
                for j, hist_entry in enumerate(level):
                    coeff_list, multi_list = hist_entry
                    mu0, sigma0 = _hist_dist(coeff_list, multi_list)
                    coeff_lj = self.WaveDec_data[l][j]
                    nsigmafit[l][j] = (coeff_lj - mu0)/sigma0
            self.Nsigma = nsigmafit
            self.NsigmaFixedRes = FixedResGlobal(self.Nsigma)
        else:
            self.PlessX, self.PeqX = self.ProbX() #Prob less extreme, Prob equally extreme
            self.Nsigma = self.NSigmaPerBin(self.PlessX) #Nsigma
            self.Log10PX = self.Log10ProbPerBin(self.PlessX, self.PeqX) #Log10 of Prob greater or equally extreme
            self.NsigmaFixedRes = FixedResGlobal(self.Nsigma)
            if extrapolate==True:
                self.PlessX_fit, self.PeqX_fit = self.Extrapolate()
                self.Nsigma_fit = self.NSigmaPerBin(self.PlessX_fit)
                self.Log10PX_fit = self.Log10ProbPerBin(self.PlessX_fit, self.PeqX_fit)
                self.NsigmaFixedRes_fit = FixedResGlobal(self.Nsigma_fit)

    @staticmethod
    def GeneratePoisson(data, seed):
        np.random.seed(seed=seed)
        pseudodata = [np.random.poisson(data[i]) for i in range(len(data))]
        return pseudodata

    def ProbX(self):
        lessX = _zeros_like(self.WaveDec_data)
        eqX = _zeros_like(self.WaveDec_data)
        for l, level in enumerate(self.WaveDec_data):
            for j, coeff in enumerate(level):
                hist = self.Histogram[l][j]
                pcoeff_list, pmulti_list = hist
                multi = 0
                if coeff in pcoeff_list:
                    index = np.where(pcoeff_list==coeff)
                    multi = pmulti_list[index]
                sum_less = 0.0
                sum_eq = 0.0
                for pmulti in pmulti_list:
                    if pmulti > multi:
                        sum_less += pmulti
                    elif pmulti == multi:
                        sum_eq += pmulti

                lessX[l][j] = float(sum_less)/float(self.Nsets)
                eqX[l][j] = float(sum_eq)/float(self.Nsets)
        return lessX, eqX

    def NSigmaPerBin(self, lessX):
        nsigma = _zeros_like(self.WaveDec_data)
        for l, level in enumerate(self.WaveDec_data):
            for j, data_coeff in enumerate(level):
                hypo_coeff = self.WaveDec_hypo[l][j]
                sign = -1 if data_coeff < hypo_coeff else 1
                nsig = float(mp.nstr(sign*_nsigma(lessX[l][j]),
                                     g_logdigits))
                nsigma[l][j] = nsig
        return nsigma

    def Log10ProbPerBin(self, lessX, eqX):
        log10probX = _zeros_like(self.WaveDec_data)
        for l, level in enumerate(self.WaveDec_data):
            for j, coeff in enumerate(level):
                grteqX = 1-lessX[l][j]
                log10pX = float(mp.nstr(mp.log10(grteqX),
                                        g_logdigits))
                log10probX[l][j] = log10pX
        return log10probX

    def Extrapolate(self):
        lessX_fit = _zeros_like(self.WaveDec_data)
        eqX_fit = _zeros_like(self.WaveDec_data)
        
        fit_params = _empty_like(self.WaveDec_data)
        fit_prob = _empty_like(self.WaveDec_data)
        for l, level in enumerate(self.Histogram):
            for j, hist_entry in enumerate(level):
                coeff_list, multi_list = hist_entry
                logmulti_list = [np.log(i) for i in multi_list]
                cmin = int(np.min(coeff_list))
                cmax = int(np.max(coeff_list))
                bounds = ([-np.inf, cmin, 0, 0, 0.8],
                          [np.inf, cmax, np.inf, np.inf, 1.3])
                mu0, sigma0 = _hist_dist(coeff_list, multi_list)
                n0 = np.log(_hist_sum(coeff_list, multi_list))-\
                            0.5*np.log(2*np.pi*sigma0**2)
                v0 = 1.0e-14 if abs(mu0)>1 else 1
                p0 = (1, mu0, sigma0, v0, 1)
                params, cov = curve_fit(_np_fitlogC, coeff_list, logmulti_list,
                               bounds=bounds, p0=p0, maxfev=3000)
                fit_params[l][j] = params.tolist()
                fit_coeffs, fit_pcoeffs = _hist_extrapolate(coeff_list, params)
                fit_prob[l][j] = [fit_coeffs, fit_pcoeffs]

        for l, level in enumerate(self.WaveDec_data):
            for j, coeff in enumerate(level):
                [c, pc] = fit_prob[l][j]
                cmin = np.min(c)
                cmax = np.max(c)
                if coeff < cmin or coeff > cmax:
                    lessX_fit[l][j] = 1-g_pmin
                    eqX_fit[l][j] = 0
                else:
                    pcoeff = 0.0
                    if coeff in c:
                        index = np.where(c==coeff)
                        pcoeff = pc[index]
                    sum = mp.mpf(0.0)
                    sumeq = mp.mpf(0.0)
                    for p in pc:
                        if p > pcoeff:
                            sum += p
                        elif p == pcoeff:
                            sumeq += p
                    lessX_fit[l][j] = sum
                    eqX_fit[l][j] = sumeq
        return lessX_fit, eqX_fit

    def zipHistogram(self, histogram):
        zipHist = _empty_like(self.WaveDec_data)
        for i, level in enumerate(histogram):
            for j, entry in enumerate(level):
                ls_zip = zip(entry[0], entry[1])
                zipHist[i][j] = list(ls_zip)
        return zipHist

def _nsigma(pLessX):
    return mp.sqrt(2)*mp.erfinv(pLessX)

def _hist_sum(coeff_list, multi_list):
    sum = 0.0
    for i in range(len(coeff_list)):
        sum += coeff_list[i]*multi_list[i]
    return sum

def _hist_dist(coeff_list, multi_list):
    sum = 0.0
    n = 0.0
    for i in range(len(coeff_list)):
        sum += coeff_list[i]*multi_list[i]
        n += multi_list[i]
    mean = float(sum)/float(n)

    var = 0.0
    for i in range(len(coeff_list)):
        var += multi_list[i]*pow(coeff_list[i]-mean, 2)
    std = np.sqrt(var/n)
    return mean, std

def _hist_extrapolate(coeff_list, params):
    coeff_max = int(np.max(coeff_list))
    coeff_min = int(np.min(coeff_list))

    n0, mu, sigma, v, p = params
    mid_coeffs = [i for i in range(coeff_min, coeff_max+1)]
    mid_pcoeffs = [_mp_fitC(i, n0, mu, sigma, v, p) for i in mid_coeffs]

    right_coeffs = []
    right_pcoeffs = []
    pc = 1
    max = coeff_max+1
    while pc > g_pmin:
        pc = _mp_fitC(max, n0, mu, sigma, v, p)
        right_coeffs.append(max)
        right_pcoeffs.append(pc)
        max += 1

    left_coeffs = []
    left_pcoeffs = []
    pc = 1
    min = coeff_min-1
    while pc > g_pmin:
        pc = _mp_fitC(min, n0, mu, sigma, v, p)
        left_coeffs.append(min)
        left_pcoeffs.append(pc)
        # c -= 1
        min -= 1

    full_coeffs = left_coeffs[::-1] + mid_coeffs + right_coeffs
    full_pcoeffs = left_pcoeffs[::-1] + mid_pcoeffs + right_pcoeffs

    hist_sum = np.sum(full_pcoeffs)

    renormize = (1-g_pmin)*mp.power(hist_sum, -1)
    renorm_pcoeffs = np.multiply(full_pcoeffs, renormize)

    return full_coeffs, renorm_pcoeffs

def _mp_fitC(c, n0, mu, sigma, v, p):
    expS = mp.mpf(-0.5 * (c - mu)**2 * pow(sigma,-2))
    expA = mp.mpf(-abs(v) * pow(abs(c), p))
    mp_fitlogC = n0 + expS + expA
    return mp.exp(mp_fitlogC)

def _np_fitlogC(c, n0, mu, sigma, v, p):
    expS = -0.5 * (c - mu)**2 * pow(sigma,-2)
    expA = -abs(v) * pow(abs(c), p)
    return n0 + expS + expA

def _np_fitC(c, n0, mu, sigma, v, p):
    expS = -0.5 * (c - mu)**2 * pow(sigma,-2)
    expA = -abs(v) * pow(abs(c), p)
    np_fitlogC = n0 + expS + expA
    return np.exp(np_fitlogC)





"""
Kawking
"""
from __future__ import absolute_import

from . import w_analysis
from . import w_transform
from . import w_plotting

from .w_analysis import *
from .w_transform import *
from .w_plotting import *

__all__ = w_analysis.__all__ + w_transform.__all__ + w_plotting.__all__



"""
Plotting Functions
Functions that generate the nsigma bar plots per level.
"""

# TODO: change parameter description to colored
from __future__ import absolute_import
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.colors import Normalize, LogNorm
from matplotlib.colorbar import ColorbarBase
from matplotlib import cm

from ..w_transform import HaarTransform, InvHaarTransform

from .plottingtools import _BinData,  _findmin, _findmax
from .plottingtools import  _NewColorMap, _NSigmaFilter

__all__ = ['nsigScalogram', 'nsigFixedRes']

data_color='#0782B0'
nsigma_color='#54B959'

ymin_delta = 0.2 # small number

def nsigScalogram(data, hypothesis, nsigma, signal_only=None,
                  nsigma_min=None, nsigma_percent=1,
                  nsigma_colorcode=False, title=None, xlabel=None, outputfile=None):
    """
    Function that plots the nsigma values of the wavelet coefficients per level.
    Parameters
    ----------
    data : array
    Array to treat as the signal in the signal reconstruction.
    hypothesis : array
    Array to treat as the background in the signal reconstruction.
    nsigma : array
    Nsigma array of the data.
    nsigma_min : float
    Minimum value of the nsigmas to use for the signal reconstruction.
    nsigma_percent : float
    Percent of the decreasing-order nsigma array to use in the signal reconstruction.
    nsigma_colorcode : array
    Array of shape (nsigma) to use as the color-code for the bins. If None, the
    right and left side of the bins will be colored slightly different.
    """
    
    nsigCcoeffs = nsigma
    Level = len(nsigCcoeffs)-1
    
    data_hist, _, data_center, data_width = _BinData(data, bins=2**(Level))
    back_hist, _, back_center, back_width = _BinData(hypothesis, bins=2**(Level))
    
    cut = '(No cut)'
    if nsigma_percent is not None:
        cut = '(Keep ' + str(nsigma_percent*100) + '%)'
    if nsigma_min is not None:
        cut = '(Sigma min = ' + str(nsigma_min)+')'
    
    DeltaCoeff = _NSigmaFilter(data, hypothesis, nsigma, nsigma_min, nsigma_percent)
    
    ReconstructedData = InvHaarTransform(DeltaCoeff, normalize=False)
    RecData = ReconstructedData
    
    nrows = Level+2
    ratio = [1.5]
    ratio += [1]*(Level+1)
    
    fig = plt.figure(figsize=(12,12))
    gs = gridspec.GridSpec(ncols=1, nrows=nrows,
                           height_ratios=ratio,
                           hspace=0)
    axs = [fig.add_subplot(gs[i,0]) for i in range(nrows)]
    
    # Fill out top panel
    axs[0].bar(data_center, data_hist, align='center',
               width=data_width, color=data_color, label="Data")

    axs[0].legend(edgecolor="black", fancybox=False, fontsize=12,
                  handlelength=0, handletextpad=0)
    axs[0].set_yscale('log')
    
    # The second panel will have the reconstructed signal
    if signal_only is not None:
        #norm = np.linalg.norm(hypothesis)
        signal_only = np.divide(signal_only, np.sqrt(hypothesis))
        signal_hist, _, signal_center, signal_width = _BinData(signal_only, bins=2**(Level))
        #axs[1].plot(signal_center, signal_hist, '*', markersize=3, color='red')
        axs[1].plot(signal_center, signal_hist, color='red', label="Generating Function")

    RecData = np.divide(RecData, np.sqrt(hypothesis))
    axs[1].plot(data_center, RecData, 'o', markersize=3, color='#E67E22', label='Reconstructed Signal {}'.format(cut))

    axs[1].plot(range(len(data_center)), np.zeros_like(RecData), color='black', linewidth=0.5)
    axs[1].set_yscale('linear')
    axs[1].legend(edgecolor="black", fancybox=False, fontsize=12)


    coeffs_min = _findmin(nsigma[:Level]) # Use to set the min ylim of the plots
    coeffs_max = _findmax(nsigma[:Level]) # Use to set the max ylim of the plots
    
    cmap = _NewColorMap()
    sig_max = _findmax(np.absolute(nsigma[:Level]))
    norm = Normalize(vmin=0, vmax=sig_max)

    for l in range(Level):
        bins = 2**(Level-l-1)
        hist, edges, center, width = _BinData(nsigma[l], bins=bins)
        midLeft = (center-edges[:-1])/2.0
        LeftCenter = edges[:-1]+midLeft
        midRight = (edges[1:]-center)/2.0
        RightCenter = center+midRight
        
        if nsigma_colorcode==True:
            norm_points = norm(np.absolute(nsigma[l]))
            color_points = [cmap(i) for i in norm_points]
            axs[l+2].bar(LeftCenter, hist, align='center', width=width/2.0,
                         color=color_points)
            axs[l+2].bar(RightCenter, hist, align='center', width=width/2.0,
                         color=color_points, alpha=0.8)
        
        else:
            axs[l+2].bar(LeftCenter, hist, align='center', width=width/2.0,
                       color='#58C85E')
            axs[l+2].bar(RightCenter, hist, align='center', width=width/2.0,
                       color='#43A961')

        axs[l+2].set_ylim(coeffs_min-ymin_delta, coeffs_max+ymin_delta)
        axs[l+2].plot(range(bins), np.zeros(bins), color='black', linewidth=0.5)
        axs[l+2].tick_params(axis='x', bottom=False, labelbottom=False)
        lev = Level-l-1
        axs[l+2].text(x=-.062, y=.66, s=r'$N\sigma(C_{l=%.1i})$'%(lev), fontsize=14,
                      #bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                      transform=axs[l+2].transAxes,
                      rotation=90)

    if nsigma_colorcode==True:
        cbar_axs = fig.add_axes([0.93, 0.15, 0.02, 0.7]) # colorbar axis
        cbar = ColorbarBase(cbar_axs, cmap=cmap, norm=norm)
    fig.suptitle("Nsigma per Level", fontsize=18, y=0.92)
    fig.text(x=0.5, y=0.1, s=xlabel, fontsize=14)
    if outputfile is not None:
        plt.savefig(outputfile)
    plt.show()




def nsigFixedRes(data, hypothesis, nsigma, nsigma_fixedres,
                 nsigma_colorcode=False, title=None, xlabel=None, outputfile=None):
    """
    Function that plots the nsigma values of the wavelet coefficients per level.
    The global significance per level is also displayed.
    Parameters
    ----------
    data : array
    Array to treat as the signal in the signal reconstruction.
    hypothesis : array
    Array to treat as the background in the signal reconstruction.
    nsigma : array
    Nsigma array of the data.
    nsigma_min : float
    Minimum value of the nsigmas to use for the signal reconstruction.
    nsigma_percent : float
    Percent of the decreasing-order nsigma array to use in the signal reconstruction.
    colorcode : array
    Array of shape (nsigma) to use as the color-code for the bins. If None, the
    right and left side of the bins will be colored slightly different.
    """
    
    nsigCcoeffs = nsigma
    Level = len(nsigCcoeffs)-1
    
    data_hist, _, data_center, data_width = _BinData(data, bins=2**(Level))
    back_hist, _, back_center, back_width = _BinData(hypothesis, bins=2**(Level))
    
    nrows = Level+1
    ratio = [1.5]
    ratio += [1]*(Level)
    
    fig = plt.figure(figsize=(12,12))
    gs = gridspec.GridSpec(ncols=2, nrows=nrows,
                           height_ratios=ratio,
                           width_ratios=[6,1],
                           hspace=0, wspace=0)
    axs = [fig.add_subplot(gs[i,0]) for i in range(nrows)]
    axs2 = [fig.add_subplot(gs[i,1]) for i in range(nrows)]
                           
    # Fill out top panel
    axs[0].bar(data_center, data_hist, align='center',
               width=data_width, color=data_color, label="Data")
    #axs[0].text(x=.94, y=.63, s='Data', fontsize=12,
    #            bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
    #            transform=axs[0].transAxes)
    axs[0].set_yscale('log')
    axs[0].legend(edgecolor="black", fancybox=False, fontsize=12,
                  handlelength=0, handletextpad=0)
    axs2[0].tick_params(axis='both', bottom=False, left=False,
                        labelbottom=False, labelleft=False)
                        
    coeffs_min = _findmin(nsigma[:Level]) # Use to set the min ylim of the plots
    coeffs_max = _findmax(nsigma[:Level]) # Use to set the max ylim of the plots
    
    cmap = _NewColorMap()
    sig_max = _findmax(np.absolute(nsigma[:Level]))
    norm = Normalize(vmin=0, vmax=sig_max)
    
    sig_min = _findmin(nsigma_fixedres[:Level])
    sig_max = _findmax(nsigma_fixedres[:Level])
                           
    for l in range(Level):
        bins = 2**(Level-l-1)
        hist, edges, center, width = _BinData(nsigma[l], bins=bins)
        midLeft = (center-edges[:-1])/2.0
        LeftCenter = edges[:-1]+midLeft
        midRight = (edges[1:]-center)/2.0
        RightCenter = center+midRight
                                   
        if nsigma_colorcode==True:
            norm_points = norm(np.absolute(nsigma[l]))
            color_points = [cmap(i) for i in norm_points]
            axs[l+1].bar(LeftCenter, hist, align='center', width=width/2.0,
                         color=color_points)
            axs[l+1].bar(RightCenter, hist, align='center', width=width/2.0,
                         color=color_points, alpha=0.8)
                                                        
        else:
            axs[l+1].bar(LeftCenter, hist, align='center', width=width/2.0, color='#58C85E')
            axs[l+1].bar(RightCenter, hist, align='center', width=width/2.0, color='#43A961')
                                                                             
        axs[l+1].set_ylim(coeffs_min-ymin_delta, coeffs_max+ymin_delta)
        axs[l+1].plot(range(bins), np.zeros(bins), color='black', linewidth=0.5)
        axs[l+1].tick_params(axis='x', bottom=False, labelbottom=False)
        lev = Level-l-1
        axs[l+1].text(x=-.075, y=.66, s=r'$N\sigma(C_{l=%.1i})$'%(lev), fontsize=14,
                      #bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                      transform=axs[l+1].transAxes, rotation=90)

        # -- Fill the right panels with the global significance per level --
        axs2[l+1].bar(0, nsigma_fixedres[l], color='#0A700F')
        axs2[l+1].tick_params(axis='x', bottom=False, labelbottom=False, labelleft=False)
        axs2[l+1].ticklabel_format(axis='y', style='sci')
        axs2[l+1].yaxis.tick_right()
        axs2[l+1].set_ylim(bottom=0.0, top=sig_max+ymin_delta)

    if nsigma_colorcode==True:
        cbar_axs = fig.add_axes([0.94, 0.15, 0.02, 0.7]) # colorbar axis
        cbar = ColorbarBase(cbar_axs, cmap=cmap, norm=norm)
    fig.suptitle(title, fontsize=18, y=0.92)
    fig.text(x=0.5, y=0.1, s=xlabel, fontsize=14)
    if outputfile is not None:
        plt.savefig(outputfile)
        plt.show()





"""
    This file contains the Exact class that performs the statistical
    calculations when nsets is None when called by waveletanalysis.py
"""
from __future__ import absolute_import
import numpy as np
import scipy.special as spf
from scipy.optimize import curve_fit
import math
from mpmath import mp
from ..w_transform import HaarTransform, wh_alj
from .analysistools import *

__all__ = ['ExactMethod']

dummy = 1000
g_nsigmamax = 10
g_pmin = mp.erfc(g_nsigmamax * mp.sqrt(0.5))
g_digits = int(math.ceil(-1 * mp.log10(g_pmin)))
g_logdigits = 17 #for nsigma and log10p output
mp.dps = g_digits + 2 #a bit extra

class ExactMethod:
    """
        Exact class contains all functions and instances used in the 'exact'
        method.
    """
    def __init__(self, data, hypothesis):
        """
        Parameters
        ----------
        data : array_like
        hypothesis : array_like
        """
        
        self.WaveDec_data = HaarTransform(data, Normalize=False)
        self.WaveDec_hypo = HaarTransform(hypothesis, Normalize=False)

        len_wdata = len(self.WaveDec_data)
        len_whypo = len(self.WaveDec_hypo)
        assert(len_wdata==len_whypo), "Data and hypothesis must have the same wavelet decomposition level."
        
        self.Level = len_wdata-1
        
        # Do the C prob first
        self.Histogram = _empty_like(self.WaveDec_data)
        self.PlessX = _zeros_like(self.WaveDec_data)
        self.PeqX = _zeros_like(self.WaveDec_data)
        for l, level in enumerate(self.WaveDec_data[:-1]):
            for j, coeff in enumerate(level):
                mu1 = self.h_mu1(hypothesis, l+1, j+1)
                mu2 = self.h_mu2(hypothesis, l+1, j+1)
                C0 = coeff
                sign = -1 if (C0>(mu1-mu2)) else 1
                p0 = self.probCmu1mu2(C0, mu1, mu2)
                C_list, p_list = [C0], [p0]
                PlessX = mp.mpf(0.0)
                C = C0+sign
                for i in range(dummy):
                    pC = self.probCmu1mu2(C, mu1, mu2)
                    if pC <= p0:
                        break
                    if sign==-1:
                        C_list.insert(0, C)
                        p_list.insert(0, pC)
                    elif sign==1:
                        C_list.append(C)
                        p_list.append(pC)
                    PlessX += pC
                    C += sign
                self.Histogram[l][j] = np.array(zip(C_list, p_list))
                if PlessX >= 1-g_pmin:
                    PlessX = 1-g_pmin
                    p0 = 0
                self.PlessX[l][j] = PlessX
                self.PeqX[l][j] = p0

        # Now fir the first trend
        mu1 = self.h_mu1(hypothesis, self.Level, 1)
        mu2 = self.h_mu2(hypothesis, self.Level, 1)
        A0 = self.WaveDec_data[-1][0]
        sign = -1 if A0 > (mu1+mu2) else 1
        p0 = self.probAmu1mu2(A0, mu1, mu2)
        A_list, p_list = [A0], [p0]
        PlessX = mp.mpf(0.0)
        A = A0 + sign
        for i in range(dummy):
            pA = self.probAmu1mu2(A, mu1, mu2)
            if pA <= p0:
                break
            if sign==-1:
                A_list.insert(0, A)
                p_list.insert(0, pA)
            elif sign==1:
                A_list.insert(-1, A)
                p_list.insert(-1, pA)
            PlessX += pA
            A+= sign
        self.Histogram[self.Level][0] = np.array(zip(A_list, p_list))
        if PlessX >= 1-g_pmin:
            PlessX = 1-g_pmin
            p0 = 0
        self.PlessX[self.Level][0] = PlessX
        self.PeqX[self.Level][0] = p0
            
        self.Nsigma = self.NSigmaPerBin(self.PlessX)
        self.NsigmaFixedRes = FixedResGlobal(self.Nsigma)

        self.zipHistogram = self.zipHistogram(self.Histogram)

    def NSigmaPerBin(self, lessX):
        nsigma = _zeros_like(self.WaveDec_data)
        for l, level in enumerate(self.WaveDec_data):
            for j, data_coeff in enumerate(level):
                hypo_coeff = self.WaveDec_hypo[l][j]
                sign = -1 if data_coeff < hypo_coeff else 1
                nsig = float(mp.nstr(sign*_nsigma(lessX[l][j]), g_logdigits))
                nsigma[l][j] = nsig
        return nsigma
    
    def zipHistogram(self, histogram):
        zipHist = _empty_like(self.WaveDec_data)
        for i, level in enumerate(histogram):
            for j, entry in enumerate(level):
                ls_zip = zip(entry[0], entry[1])
                zipHist[i][j] = list(ls_zip)
        return zipHist


    @staticmethod
    def h_mu1(array, l, j):
        return wh_alj(array, l-1, 2*j-1)
    
    @staticmethod
    def h_mu2(array, l, j):
        return wh_alj(array, l-1, 2*j)

    @staticmethod
    def probAmu1mu2(A, mu1, mu2):
        logAfact = mp.log(mp.gamma(A+1))
        log_probAmu1mu2 = -(mu1+mu2) + A*mp.log(mu1+mu2) - logAfact
        return mp.exp(log_probAmu1mu2)

    @staticmethod
    def probCmu1mu2(C, mu1, mu2):
        v, z = abs(C), 2*mp.sqrt(mu1*mu2)
        log_BesselI = mp.log(mp.besseli(v, z, maxterms={}))
        log_probCmu1mu2 = -mu1 - mu2 + 0.5*C*(mp.log(mu1) - mp.log(mu2)) + log_BesselI
        return mp.exp(log_probCmu1mu2)


def _nsigma(pLessX):
    return mp.sqrt(2)*mp.erfinv(pLessX)





from __future__ import absolute_import
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from ..w_transform import HaarTransform, InvHaarTransform

def _zeros_like(obj):
    zeros = [np.zeros_like(lev, dtype=float) for lev in obj]
    return zeros


__all__ = ['_findmin', '_findmax', '_BinData', '_NewColorMap',
           '_NSigmaFilter']

delta = 0.0 # Small number

def _findmin(array):
    minn = delta
    for i in array:
        if np.min(i) < minn:
            minn = np.min(i)
    return minn

def _findmax(array):
    maxx = delta
    for i in array:
        if np.max(i) > maxx:
            maxx = np.max(i)
    return maxx

def _BinData(data, bins):
    hist, edges = np.histogram(a=range(bins), bins=bins, weights=data)
    center = (edges[:-1]+edges[1:])/2.0
    width = edges[1:]-edges[:-1]
    return hist, edges, center, width

def _NewColorMap():
    R=float(0+172+242)
    G=(41.+181.+104.)
    B=(242.+81.+59.)
    #colors = [(0, 0, 1), (0, 1, 0), (1, 0, 0)] #RGB
    #colors = [(0.172, 0.521, 0.729), (0.870, 0.325, 0.129)]
    colors = [(0.152, 0.552, 0.607),
              (0.666, 0.882, 0.035),
              (0.945, 0.337, 0.074)]
    nbins=2**15
    cmap_name='New'
    cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=nbins)
    return cm


def _NSigmaFilter(data, hypothesis, nsigma,
                  nsigma_min=None, nsigma_percent=None):

    WaveDec_data = HaarTransform(data)
    DataCoeffs = WaveDec_data[:-1]
    DataFirstTrend = WaveDec_data[-1]
    
    WaveDec_hypo = HaarTransform(hypothesis)
    HypoCoeffs = WaveDec_hypo[:-1]
    HypoFirstTrend = WaveDec_hypo[-1]

    Level = len(DataCoeffs)

    flatNsigma = []
    flatAbsNsigma = []
    flatDataCoeffs = []
    flatHypoCoeffs = []
    flatLoc = []

    count = 0
    for l in range(Level):
        J = 2**(Level-l-1)
        for j in range(J):
            flatNsigma.append(nsigma[l][j])
            flatAbsNsigma.append(abs(nsigma[l][j]))
            flatDataCoeffs.append(DataCoeffs[l][j])
            flatHypoCoeffs.append(HypoCoeffs[l][j])
            flatLoc.append([l, j])
            count += 1

    ixsort = np.argsort(flatAbsNsigma)[::-1]
    sortNsigma = [flatNsigma[ix] for ix in ixsort]
    sortDataCoeffs = [flatDataCoeffs[ix] for ix in ixsort]
    sortHypoCoeffs = [flatHypoCoeffs[ix] for ix in ixsort]
    sortLoc = [flatLoc[ix] for ix in ixsort]

    keepNsigma = []
    keepDeltaCoeff = []
    keepLoc = []
    if nsigma_min is not None:
        for i in range(len(sortNsigma)):
            if abs(sortNsigma[i]) > nsigma_min:
                keepNsigma.append(sortNsigma[i])
                keepDeltaCoeff.append(sortDataCoeffs[i]-sortHypoCoeffs[i])
                keepLoc.append(sortLoc[i])
                
    elif nsigma_percent is not None:
        net = len(sortNsigma)
        netkeep = int(np.ceil(net*nsigma_percent))
        keepNsigma = sortNsigma[:netkeep]
        keepDeltaCoeff = np.subtract(sortDataCoeffs[:netkeep],
                                     sortHypoCoeffs[:netkeep])
        keepLoc = sortLoc[:netkeep]

    else:
        keepNsigma = sortNsigma
        keepDeltaCoeff = np.subtract(sortDataCoeffs,
                                     sortHypoCoeffs)
        keepLoc = sortLoc

    keep = _zeros_like(WaveDec_data)
    for i in range(len(keepDeltaCoeff)):
        l = keepLoc[i][0]
        j = keepLoc[i][1]
        keep[l][j] = keepDeltaCoeff[i]
    keep[-1][0] = DataFirstTrend-HypoFirstTrend

    return keep






"""
Plotting Functions
Functions that generate the wavelet scalograms.
"""
from __future__ import absolute_import
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.colors import Normalize, LogNorm
from matplotlib.colorbar import ColorbarBase
from matplotlib import cm

from ..w_transform import HaarTransform

from .plottingtools import _BinData, _findmin, _findmax
from .plottingtools import _NewColorMap

__all__ = ['wScalogram', 'wScalogram_nsig']

Data_color='#0782B0'
Coeffs_color='#69B4F2'
Firsttrend_color=Coeffs_color
Nsigma_color='#54B959'

def wScalogram(data, hypothesis=None, signal_only=None, firsttrend=False, logscale=True,
               filled=False, title=None, xlabel=None, outputfile=None):
    """
    Function that generates a bar plot of the wavelet coefficients of the data array
    per level.
    Parameters
    ----------
    data : array
    Array to calculate the discrete Haar wavelet transform on.
    firsttrend : bool
    Whether to include the first trend on the scalogram plot.
    filled : bool
    Whether to fill the bars or just show their contour.
    outputfile : string
    Name of the png file to save the plot to. If None, don't print the plot.
    """
    
    WaveDec_data = HaarTransform(data)
    Ccoeffs = WaveDec_data[:-1]
    FirstTrend = WaveDec_data[-1]
    Level = len(Ccoeffs)
    
    nlevels = Level if firsttrend==False else Level+1
    nrows = nlevels+1 # the first panel is the data histogram
    if signal_only is not None:
        nrows += 1 # add another panel for the generating function
    ratio = [1.5]
    ratio += [1]*(nrows-1)
    
    if filled==True:
        histtype='bar'
        coeffs_color=Coeffs_color
        firsttrend_color=Firsttrend_color
    else:
        histtype='step'
        coeffs_color='black'
        firsttrend_color='black'

    if logscale==True:
        scale='log'
    else:
        scale='linear'

    fig = plt.figure(figsize=(12,12))
    gs = gridspec.GridSpec(ncols=1, nrows=nrows,
                           height_ratios=ratio,
                           hspace=0)
    axs = [fig.add_subplot(gs[i,0]) for i in range(nrows)]

    # Fill out top panel
    data_hist, _, data_center, data_width = _BinData(data, bins=2**Level)
    axs[0].bar(data_center, data_hist, align='center',
               width=data_width, color=Data_color)
    axs[0].text(x=.93, y=.63, s='Data', fontsize=12,
                bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                transform=axs[0].transAxes)
    axs[0].set_yscale(scale)

    # If generating function is provided
    if signal_only is not None:
        if hypothesis is not None:
            signal_only = np.divide(signal_only, np.sqrt(hypothesis))
        signal_hist, _, signal_center, signal_width = _BinData(signal_only, bins=2**(Level))
        axs[1].plot(signal_center, signal_hist, color='red', label="Generating Function")
        axs[1].tick_params(axis='both', bottom=False, labelbottom=False)
        axs[1].set_yscale('linear')
        axs[1].legend(edgecolor="black", loc=2, fancybox=False, fontsize=12)
    
    # If firsttrend, fill out the bottom panel with the first trend
    if firsttrend==True:
        bins = 1
        axs[-1].hist(x=range(bins), bins=bins, weights=FirstTrend,
                     histtype=histtype, color=firsttrend_color)
        axs[-1].tick_params(axis='both', bottom=False, labelbottom=False)
        axs[-1].set_yscale(scale)
        axs[-1].text(x=.93, y=.63, s=r'$A_{l=%.1i}$'%(0), fontsize=12,
                     bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                     transform=axs[-1].transAxes)

    # Fill out the rest of the pannels with the wavelet coefficients
    # If signal_only, start two panels below the top panel
    s = 2 if signal_only is not None else 1
    for l in range(Level):
        bins=2**(Level-l-1)
        coeffs = Ccoeffs[l]
        
        if logscale==True:
            # Plot the positive coefficients
            pos_ix = np.where(Ccoeffs[l]>0)
            pos_coeffs = np.zeros_like(coeffs)
            for i in pos_ix:
                pos_coeffs[i] = coeffs[i]
            axs[l+s].hist(x=range(bins), bins=bins,
                          weights=pos_coeffs, histtype=histtype, color=coeffs_color)

            # Now plot the negative coefficients. The bars are hashed to distinguish the
            # pos and neg coefficients.
            neg_ix = np.where(Ccoeffs[l]<0)
            neg_coeffs = np.zeros_like(coeffs)
            for j in neg_ix:
                neg_coeffs[j] = np.absolute(coeffs[j])
            axs[l+s].hist(x=range(bins), bins=bins,
                          weights=neg_coeffs, histtype=histtype, hatch='///', color=coeffs_color)

            axs[l+s].tick_params(axis='both', bottom=False, labelbottom=False)
            lev = Level-l-1
            axs[l+s].text(x=.93, y=.63, s=r'$C_{l=%.1i}$'%(lev), fontsize=12,
                          bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                          transform=axs[l+1].transAxes)
            axs[l+s].set_yscale(scale)

        else:
            axs[l+s].hist(x=range(bins), bins=bins, weights=coeffs, histtype=histtype, color=coeffs_color)
            axs[l+s].tick_params(axis='both', bottom=False, labelbottom=False)
            lev = Level-l-1
            axs[l+s].text(x=.93, y=.63, s=r'$C_{l=%.1i}$'%(lev), fontsize=12,
                          bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                          transform=axs[l+1].transAxes)
            axs[l+s].set_yscale(scale)

    fig.suptitle(title, fontsize=18, y=0.92)
    fig.text(x=0.5, y=0.1, s=xlabel, fontsize=14)
    if outputfile is not None:
        plt.savefig(outputfile)
    plt.show()


def wScalogram_nsig(data, hypothesis=None, signal_only=None, nsigma=None,
                    firsttrend=False, logscale=True,
                    title=None, xlabel=None, outputfile=None):
    """
    Function that generates a bar plot of the wavelet coefficients of the data array
    per level.
    Parameters
    ----------
    data : array
    Array to calculate the discrete Haar wavelet transform on.
    nsigma : array
    Nsigma array to use as the color-code for the wavelet coefficients.
    firsttrend : bool
    Whether to include the first trend on the scalogram plot.
    logscale : bool
    Whether to use a linear of log scale on the y-axis .
    outputfile : string
    Name of the png file to save the plot to. If None, don't print the plot.
    """
    
    WaveDec_data = HaarTransform(data)
    Ccoeffs = WaveDec_data[:-1]
    FirstTrend = WaveDec_data[-1]
    Level = len(Ccoeffs)
    
    if logscale==True:
        scale='log'
    else:
        scale='linear'

    nlevels = Level if firsttrend==False else Level+1
    nrows = nlevels+1 # the first panel is the data histogram
    if signal_only is not None:
        nrows += 1 # add another panel for the generating function
    ratio = [1.5]
    ratio += [1]*(nrows-1)

    fig = plt.figure(figsize=(12,12))
    gs = gridspec.GridSpec(ncols=1, nrows=nrows,
                           height_ratios=ratio,
                           hspace=0)
    axs = [fig.add_subplot(gs[i,0]) for i in range(nrows)]
    cbar_axs = fig.add_axes([0.93, 0.15, 0.02, 0.7]) # colorbar axis
                           
    # Fill out top panel
    data_hist, _, data_center, data_width = _BinData(data, bins=2**Level)
    axs[0].bar(data_center, data_hist, align='center', width=data_width, color=Data_color)
    axs[0].set_yscale(scale)
    axs[0].text(x=.93, y=.63, s='Data', fontsize=12,
                bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                transform=axs[0].transAxes)
                
    # If generating function is provided
    if signal_only is not None:
        if hypothesis is not None:
            signal_only = np.divide(signal_only, np.sqrt(hypothesis))
        signal_hist, _, signal_center, signal_width = _BinData(signal_only, bins=2**(Level))
        axs[1].plot(signal_center, signal_hist, color='red', label="Generating Function")
        axs[1].tick_params(axis='both', bottom=False, labelbottom=False)
        axs[1].set_yscale('linear')
        axs[1].legend(edgecolor="black", loc=2, fancybox=False, fontsize=12)

    cmap = _NewColorMap()
    binintensity = np.absolute(nsigma)
    vmin = _findmin(binintensity)
    vmax = _findmax(binintensity)
    norm = Normalize(vmin=vmin, vmax=vmax)
                  
    # If firsttrend, fill out the bottom panel with the first trend
    if firsttrend==True:
        bins=1
        norm_points = norm(binintensity[-1])
        color_points = [cmap(i) for i in norm_points]
        hist, _, center, width = _BinData(FirstTrend, bins=1)
        axs[-1].bar(center, hist, align='center', width=width, color=color_points)
        axs[-1].tick_params(axis='both', bottom=False, labelbottom=False)
        axs[-1].set_yscale(scale)
        axs[-1].text(x=.93, y=.63, s=r'$A_{l=%.1i}$'%(0), fontsize=12,
                     bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                     transform=axs[-1].transAxes)

    # Now plot the negative coefficients. The bars are hashed to distinguish the
    # pos and neg coefficients.
    s = 2 if signal_only is not None else 1
    for l in range(Level):
        bins=2**(Level-l-1)
        coeffs = Ccoeffs[l]
        norm_points = norm(binintensity[l])
        color_points = [cmap(i) for i in norm_points]
        
        if logscale==True:
            # Plot the positive coefficients
            pos_ix = np.where(coeffs>0)
            pos_coeffs = np.zeros_like(coeffs)
            for i in pos_ix:
                pos_coeffs[i] = coeffs[i]
            pos_hist, _, pos_center, pos_width = _BinData(pos_coeffs, bins=bins)
            axs[l+s].bar(pos_center, pos_hist, align='center', width=pos_width, color=color_points)
             
            # Now plot the negative coefficients. The bars are hashed to distinguish the
            # pos and neg coefficients.
            neg_ix = np.where(Ccoeffs[l]<0)
            neg_coeffs = np.zeros_like(coeffs)
            for j in neg_ix:
                neg_coeffs[j] = np.absolute(coeffs[j])
            neg_hist, _, neg_center, neg_width = _BinData(neg_coeffs, bins=bins)
            axs[l+s].bar(neg_center, neg_hist, align='center', width=neg_width, color=color_points,
                         hatch='///')
                                   
            axs[l+s].tick_params(axis='both', bottom=False, labelbottom=False)
            lev = Level-l-1
            axs[l+s].text(x=.93, y=.63, s=r'$C_{l=%.1i}$'%(lev), fontsize=12,
                          bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                          transform=axs[l+1].transAxes)
            axs[l+s].set_yscale(scale)
        
        else:
            hist, _, center, width = _BinData(coeffs, bins=bins)
            axs[l+s].bar(center, hist, align='center', width=width,
                         color=color_points)
            axs[l+s].plot(range(bins), np.zeros(bins), color='black',
                          linewidth=0.5)
            axs[l+s].tick_params(axis='both', bottom=False, labelbottom=False)
            lev=Level-l-1
            axs[l+s].text(x=.93, y=.63, s=r'$C_{l=%.1i}$'%(lev), fontsize=12,
                          bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 5},
                          transform=axs[l+1].transAxes)
            axs[l+s].set_yscale(scale)

    cbar = ColorbarBase(cbar_axs, cmap=cmap, norm=norm)
    fig.suptitle(title, fontsize=18, y=0.92)
    fig.text(x=0.5, y=0.1, s=xlabel, fontsize=14)
    if outputfile is not None:
        plt.savefig(outputfile)
    plt.show()



"""
Plotting Functions
Functions that generate the bin scalograms of the coefficeints
and nsigma values.
"""
from __future__ import absolute_import
from . import plotting

__all__ = ['wScalogram', 'wScalogram_nsig',
           'nsigScalogram', 'nsigFixedRes']
           
wScalogram = plotting.wScalogram
wScalogram_nsig = plotting.wScalogram_nsig

nsigScalogram = plotting.nsigScalogram
nsigFixedRes = plotting.nsigFixedRes




from setuptools import setup, find_packages

setup(
    name = 'kwak',
    packages=find_packages(),
    version='1.1.0',
    url='https://github.com/alexxromero/kwak_wavelets',
    author='Ben G. Lillard and Alexis Romero',
    author_email='alexir2@uci.edu',
    license='MIT',
    description='Statistical tool for wavelet analysis',
    )



import numpy as np
from mpmath import mp

__all__ = ['_empty_like', '_zeros_like', 'FixedResGlobal']

def _empty_like(obj):
    empty = [np.empty_like(lev, dtype=object) for lev in obj]
    return empty

def _zeros_like(obj):
    zeros = [np.zeros_like(lev, dtype=float) for lev in obj]
    return zeros

def FixedResGlobal(nsigma):
    n_levels = len(nsigma)
    q_perlevel = np.empty(n_levels, dtype=object)
    for l, lev in enumerate(nsigma):
        q_total = 0
        nQs = len(lev)
        for j, sig in enumerate(lev):
            pval = mp.erfc(abs(sig)/mp.sqrt(2))
            q = -2*mp.log(pval)
            q_total += q
        q_perlevel[l] = [q_total, nQs]

    nsigma_fixedres = np.zeros(n_levels)
    for i, entry in enumerate(q_perlevel):
        qT = entry[0]
        nQ = entry[1]
        Dchi2 = mp.gammainc(nQ, 0, 0.5*qT)/mp.gamma(nQ)
        nsigma_ell = mp.sqrt(2)*mp.erfinv(Dchi2)
        nsigma_fixedres[i] = nsigma_ell
    return nsigma_fixedres



"""
This file contains the classes nsets and exact that compute the
statistical analysis of the data based on the given hypothesis array.
TODO: fastgaussian+extrapolate asserts
"""
from __future__ import absolute_import
import os
import numpy as np
import math
import pandas as pd

from .w_transform import HaarTransform
from .tools import NsetsMethod, ExactMethod

__all__ = ['nsets', 'exact']

class nsets:
    def __init__(self, data, hypothesis, nsets, seed=None, extrapolate=False,
                 fastGaussian=False, outputdir=None):
        """
        Compute the statistical analysis for the data by sampling the
        hypothesis from a Poisson distribution.
        Parameters
        ----------
        ::data:: int array_like
          Binned distribution of the data.
        ::hypothesis:: array_like
          Binned background distribution.
        ::nsets:: int
          Number of hypothesis-like arrays to sample from a Poisson
          distribution. If nsets is None, the 'exact' method will be used.
        ::extrapolate:: bool
          If extrapolate is True, fit a curve to the nset data.
          Only possible if nsets is not None.
        ::fastGaussian:: boolean
          If True, calculates Nsigma and NsigmaFRGS directly from the
          mean and standard deviation. Does not reconstruct the probability
          distribution for wavelet coefficients.
        ::outputdir:: string
          File to save all instances to.
        """
        
        data = np.asarray(data)
        hypothesis = np.asarray(hypothesis)
        
        data_type = type(data[0].item()) # Input data must be integer type
        assert(issubclass(data_type, int)), "Data array must be integer valued."
        assert(len(data)==len(hypothesis)), "Data and hypothesis arrays must have the same length."
        
        # -- Argument options ----------
        self.nsets = nsets
        self.seed = seed
        self.extrapolate = extrapolate
        self.fast = fastGaussian
        # ------------------------------

        self.WaveDec_data = HaarTransform(data, Normalize=False)
        self.WaveDec_hypothesis = HaarTransform(hypothesis, Normalize=False)

        NsetsAnalysis = NsetsMethod(data, hypothesis, nsets, extrapolate, fastGaussian, seed)
        self.Level = NsetsAnalysis.Level # Max level of the discrete Haar wavelet transformation of the data
        self.Histogram = NsetsAnalysis.zipHistogram # List of coefficients per level and their multipicity (coeff, multi)
        self.Nsigma = NsetsAnalysis.Nsigma # Nsigma per coefficient
        #self.PlessX = NsetsAnalysis.PlessX # Prob. of obtaining a less extreme coeff. value
        #self.PeqX = NsetsAnalysis.PeqX # Prob. of obtaining an equally extreme coeff. value
        self.NsigmaFixedRes = NsetsAnalysis.NsigmaFixedRes # Global significance of Nsigma per level
        
        # -- If fastGaussian is False --
        if self.fast==False:
            self.PlessX = NsetsAnalysis.PlessX
            self.PeqX = NsetsAnalysis.PeqX
            # -- If extrapolate is True --
            if extrapolate==True:
                self.Nsigma_fit = NsetsAnalysis.Nsigma_fit
                self.PlessX_fit = NsetsAnalysis.PlessX_fit
                self.PeqX_fit = NsetsAnalysis.PeqX_fit
                self.NsigmaFixedRes_fit = NsetsAnalysis.NsigmaFixedRes_fit

        # -- Write output to csv files --
        # -- outputdir must be unique --
        if outputdir is not None:
            path = os.getcwd() + "/" + outputdir
            print("Printing files")
            os.mkdir(path)
            
            self.printInfo(path)
            Nsigma_dframe = _DataFrame(self.Nsigma)
            PlessX_dframe = _DataFrame(self.PlessX)
            PeqX_dframe = _DataFrame(self.PeqX)
            FixedRes_dframe = _scalarDataFrame(self.NsigmaFixedRes)
            
            Nsigma_dframe.to_csv(path+"/Nsigma.csv", index=False)
            PlessX_dframe.to_csv(path+"/PlessX.csv", index=False)
            PeqX_dframe.to_csv(path+"/PeqX.csv", index=False)
            FixedRes_dframe.to_csv(path+"/NsigmaFixedRes.csv", index=False)
            
            if self.fast==False:
                PlessX_dframe = _DataFrame(self.PlessX)
                PeqX_dframe = _DataFrame(self.PeqX)
                
                PlessX_dframe.to_csv(path+"/PlessX.csv", index=False)
                PeqX_dframe.to_csv(path+"/PeqX.csv", index=False)
            
                if extrapolate==True:
                    Nsigma_dframe_fit = _DataFrame(self.Nsigma_fit)
                    PlessX_dframe_fit = _DataFrame(self.PlessX_fit)
                    PeqX_dframe_fit = _DataFrame(self.PeqX_fit)
                    FixedRes_dframe_fit = _scalarDataFrame(self.NsigmaFixedRes_fit)

                    Nsigma_dframe_fit.to_csv(path+"/Nsigma_fit.csv", index=False)
                    PlessX_dframe_fit.to_csv(path+"/PlessX_fit.csv", index=False)
                    PeqX_dframe_fit.to_csv(path+"/PeqX_fit.csv", index=False)
                    FixedRes_dframe_fit.to_csv(path+"/NsigmaFixedRes_fit.csv", index=False)

    def printInfo(self, path):
        f = open(path+"/Info.txt", 'w')
        f.write("Method : nsets")
        f.write("\n")
        f.write("Wavelet Decomposition Level : {} ".format(self.Level))
        f.write("\n")
        f.write("Np.sets : {}".format(self.nsets))
        f.write("\n")
        f.write("Fast Version (Gaussian) : {}".format(self.fast))
        f.write("\n")
        f.write("Extrapolate : {}".format(self.extrapolate))
        f.close()



class exact:
    def __init__(self, data, hypothesis, outputdir=None):
        """
        Compute the statistical analysis for the data given a hypothesis.
        Parameters
        ----------
        ::data:: int array_like
          Binned distribution of the data.
        ::hypothesis:: array_like
          Binned background distribution.
        ::outputdir:: string
          File to save all instances to.
        """
        
        data = np.asarray(data)
        hypothesis = np.asarray(hypothesis)
        
        data_type = type(data[0]) # Input data must be integer type
        assert(issubclass(data_type, int)), "Data array must be integer valued."

        self.WaveDec_data = HaarTransform(data, Normalize=False)
        self.WaveDec_hypothesis = HaarTransform(hypothesis, Normalize=False)

        ExactAnalysis = ExactMethod(data, hypothesis)
        self.Level = ExactAnalysis.Level # Max level of the discrete Haar wavelet transformation of the data
        self.Histogram = ExactAnalysis.zipHistogram # List of coefficients per level and their multipicity (coeff, multi)
        self.Nsigma = ExactAnalysis.Nsigma # Nsigma per coefficient
        self.PlessX = ExactAnalysis.PlessX # Prob. of obtaining a less extreme coeff. value
        self.PeqX = ExactAnalysis.PeqX # Prob. of obtaining an equally extreme coeff. value
        self.NsigmaFixedRes = ExactAnalysis.NsigmaFixedRes # Global significance of Nsigma per level

        # -- Write output to csv files --
        # -- outputdir must be unique --
        if outputdir is not None:
            path = os.getcwd() + "/" + outputdir
            print("Printing files")
            os.mkdir(path)
        
            #_nsetsInfo(path)
            Nsigma_dframe = _DataFrame(self.Nsigma)
            PlessX_dframe = _DataFrame(self.PlessX)
            PeqX_dframe = _DataFrame(self.PeqX)
            FixedRes_dframe = _scalarDataFrame(self.NsigmaFixedRes)
            
            Nsigma_dframe.to_csv(path+"/Nsigma.csv", index=False)
            PlessX_dframe.to_csv(path+"/PlessX.csv", index=False)
            PeqX_dframe.to_csv(path+"/PeqX.csv", index=False)
            FixedRes_dframe.to_csv(path+"/NsigmaFixedRes.csv", index=False)


def _DataFrame(obj):
    dframes = []
    lev = len(obj)-1
    for i, lev_array in enumerate(obj[:-1]):
        label = "C"+str(lev-i-1)
        data = lev_array
        
        df = pd.DataFrame({label : data})
        dframes.append(df)
    
    label = "A0"
    data = obj[-1]
    
    df = pd.DataFrame({label : data})
    dframes.append(df)

    concat_dframe = pd.concat(dframes, axis=1)
    
    return concat_dframe


def _scalarDataFrame(obj):
    dframes = []
    lev = len(obj)-1
    for i, lev_array in enumerate(obj[:-1]):
        label = "C"+str(lev-i-1)
        data = lev_array
        
        df = pd.DataFrame({label : [data]})
        dframes.append(df)
    
    label = "A0"
    data = obj[-1]
    
    df = pd.DataFrame({label : [data]})
    dframes.append(df)

    concat_dframe = pd.concat(dframes, axis=1)

    return concat_dframe

